# Base for Feature and its non-concrete subclasses
#
# Copyright (c) 2018 Ensoft Ltd

import logging, sys
from entrance.connection.base import ConState

log = logging.getLogger(__name__)

class Feature():
    """
    Base class for a plugin that handles some set of Websocket request types.
    Classes should not subclass this directly, but instead one of the
    ConfiguredFeature or DynamicFeature subclasses.
    """

    # Name for the feature - concrete subclasses must have one
    name = None

    # Requests (with arguments) that a subclass accepts
    requests = {}

    # Notifications that a subclass sends. None can be used as an RPC reply
    notifications = [None]

    def __init__(self, ws_handler):
        self.ws_handler = ws_handler

    def close(self):
        """
        Websocket has closed
        """
        pass

    @classmethod
    def normalize_schema(feature_cls):
        """
        Ensure that each Feature subclass's schema includes any requests and
        notifications from its parent classes. In addition, add 'None' as a
        permissible notification type, since this is used for RPC replies, and
        automatically permit each req_type as a nfn_type (for use in replies)
        """
        def normalize_cls(current_cls, parent_cls):
            current_cls.requests = {**current_cls.requests, **parent_cls.requests}
            current_cls.notifications = frozenset(
                set(current_cls.notifications) |
                set(parent_cls.notifications) |
                current_cls.requests.keys()
            )
            for child_cls in current_cls.__subclasses__():
                normalize_cls(child_cls, current_cls)

        for cls in feature_cls.__subclasses__():
            normalize_cls(cls, feature_cls)

    async def handle(self, req):
        """
        Handle incoming websocket requests
        """
        try:
            # Extract out the request type and the arguments it wants. A magic
            # argument name of '__req__' means supply the entire request, and
            # and argument name starting with '?' is optional (value None if
            # unspecified)
            req_type = req['req_type']
            arg_names = self.requests[req_type]
            def get_arg(arg):
                if arg == '__req__':
                    return req
                elif arg.startswith('?'):
                    return req.get(arg[1:], None)
                else:
                    return req[arg]
            args = [get_arg(arg) for arg in arg_names]
        except KeyError:
            # Couldn't extract the request type and arguments as per the
            # supplied schema. Can't proceed.
            log.info('Received an unparseable JSON request')
            await self._notify(nfn_type='error', channel='error',
                               value='Unparseable JSON request: {}'.format(req))
            return

        # Do the operation
        try:
            result = await getattr(self, 'do_' + req_type)(*args)
        except Exception as e:
            msg = 'Exception handling {}({})'.format(req_type, ', '.join(args))
            log.error(msg + ' (see debug.log for details)')
            log.debug('Exception details', exc_info=True, stack_info=True)
            result = self._rpc_failure(msg)

        # Return a notification with the result if provided
        if result is not None:
            # Reflect back the mandatory "channel" and optional "id" from the
            # client, plus if the nfn_type is missing then insert the req_type
            # (as is typical for eg RPCs)
            result['channel'] = req['channel']
            if 'id' in req:
                result['id'] = req['id']
            if 'target' in req:
                result['target'] = req['target']
            if result.get('nfn_type', None) is None:
                result['nfn_type'] = req_type
            await self.ws_handler.notify(**result)

    def _check_nfn_type(self, nfn_type):
        """
        Check that we're conforming to our own schema declaration
        """
        if nfn_type not in self.notifications:
            log.critical('\n!!\n!!\n!! Feature {} trying to send disallowed '
                         'notification {}\n!!\n!!'.format(self.name, nfn_type))
            sys.exit(1)

    async def _notify(self, **nfn):
        """
        Helper method to send an outgoing notification. Can be overidden by
        subclasses (eg to ensure additional fields are set)
        """
        self._check_nfn_type(nfn['nfn_type'])
        await self.ws_handler.notify(**nfn)

    def _result(self, nfn_type, value=None, **kwargs):
        """
        Compose a result dict. Always needs a nfn_type, quite often has a value,
        and otherwise has other key-value pairs.
        """
        self._check_nfn_type(nfn_type)
        kwargs['nfn_type'] = nfn_type
        if value is not None:
            kwargs['value'] = value
        return kwargs

    def _rpc_success(self, result="", **kwargs):
        """
        Compose a result dict for an RPC reply after success
        """
        # nfn_type of None means use the original req_type
        return self._result(nfn_type=None, result=result, **kwargs)

    def _rpc_failure(self, error="", **kwargs):
        """
        Compose a result dict for an RPC reply after failure
        """
        # nfn_type of None means use the original req_type
        return self._result(nfn_type=None, error=error, **kwargs)



# Feature implementing core functionality
#
# Copyright (c) 2018 Ensoft Ltd

import logging, os
from .cfg_base import ConfiguredFeature
from .dyn_base import DynamicFeature
from .tgt_base import TargetFeature

log = logging.getLogger(__name__)

class CoreFeature(ConfiguredFeature):
    """
    Feature to manage core services
    """
    #
    # Schema
    #
    name = 'core'

    requests = {'force_restart': [],
                'ping': [],
                'start_feature': ['feature', 'channel', 'target', '__req__'],
                'stop_feature': ['feature', 'channel', 'target']}

    notifications = ['pong']

    config = {
        'allow_restart_requests': False,
        'allowed_dynamic_features': None
    }

    #
    # Implementation
    #
    def __init__(self, ws_handler, config):
        super().__init__(ws_handler, config)
        self.started_features = {}

    async def do_force_restart(self):
        """
        Forcibly restart the server (assuming started by the "run" script)
        """
        if self.config['allow_restart_requests']:
            os._exit(42)
        else:
            return self._rpc_failure('Restart disallowed by configuration')

    async def do_ping(self):
        """
        Respond to a keepalive message from the client
        """
        return self._result('pong')

    async def do_start_feature(self, feature_name, channel, target, req):
        """
        Start an optional feature
        """
        allowed = self.config['allowed_dynamic_features']
        if allowed is not None and feature_name not in allowed:
            return self._rpc_failure(
                'Feature {} is disallowed by configuration'.format(feature_name))
        feature_cls = DynamicFeature.find(feature_name)
        new_feature = feature_cls(self.ws_handler, channel, target, req)
        self.ws_handler.add_feature(new_feature, channel, target)
        self.started_features[_fk(feature_name, channel, target)] = new_feature

    async def do_stop_feature(self, feature_name, channel, target):
        """
        Stop an optional feature
        """
        feature_key = _fk(feature_name, channel, target)
        feature = self.started_features[feature_key]
        await self.ws_handler.remove_feature(feature, channel, target)
        try:
            del self.started_features[feature_key]
        except KeyError:
            # probably some race condition
            log.debug('Ignoring missing started feature {} on stop_feature request'.format(feature_key))
        if isinstance(feature, TargetFeature):
            # Might as well try a disconnect
            log.debug('About to disconnect stopped feature ' + feature_key)
            await feature.disconnect()

def _fk(feature_name, channel, target):
    """Construct a dict key for a feature instance"""
    return '{}::{}::{}'.format(feature_name, channel, target)



# Base class for a Netconf ThreadedConnection
#
# Copyright (c) 2018 Ensoft Ltd

from entrance.connection.threaded import ThreadedConnection

class ThreadedNCConnection(ThreadedConnection):
    """
    Base class for a ThreadedConnection whose worker thread
    is a ncclient.manager session. Just add a connection.
    """

    async def get(self, xml_filter, override=False):
        """
        Issue a Netconf "get" request in the worker thread
        """
        return await self._request('get', override, xml_filter)

    async def get_config(self, xml_filter, override=False):
        """
        Issue a Netconf "get-config" request in the worker thread
        """
        if xml_filter.strip() == '':
            xml_filter = None
        return await self._request('get_config', override, xml_filter)

    async def edit_config(self, xml_config, override=False):
        """
        Issue a Netconf "edit-config" request in the worker thread
        """
        return await self._request('edit_config', override, xml_config)

    async def commit(self, override=False):
        """
        Issue a Netconf "commit" request in the worker thread
        """
        return await self._request('commit', override)

    async def validate(self, override=False):
        """
        Issue a Netconf "validate" request in the worker thread
        """
        return await self._request('validate', override)

    async def discard_changes(self, override=False):
        """
        Issue a discard-changes request in the worker thread
        """
        return await self._request('discard_changes', override)

    def _handle_get(self, xml_filter):
        """
        Netconf "get" request
        """
        return self.mgr.get(filter=xml_filter)

    def _handle_get_config(self, xml_filter):
        """
        Netconf "get-config" request
        """
        return self.mgr.get_config(source='running', filter=xml_filter)

    def _handle_edit_config(self, xml_config):
        """
        Netconf "edit-config" request
        """
        self._handle_discard_changes()
        return self.mgr.edit_config(target='candidate', config=xml_config)

    def _handle_commit(self):
        """
        Netconf "commit" request
        """
        return self.mgr.commit()

    def _handle_validate(self):
        """
        Netconf "validate" request
        """
        return self.mgr.validate(source='candidate')

    def _handle_discard_changes(self):
        """
        Discard current changes
        """
        return self.mgr.discard_changes()



# Base class that represents a protocol sesssion owned by a separate
# synchronous thread
#
# Copyright (c) 2018 Ensoft Ltd

import logging, re, threading, traceback
import asyncio, janus

from .base import Connection, ConState
from .._util import events

log = logging.getLogger(__name__)

class ConnectionError(Exception):
    pass

class ThreadedConnection(Connection):
    """
    Base class representing a Connection where there is a dedicated thread to
    interact with the peer, which is an implementation choice that is hidden
    from the regular asyncio API as per all other Connection types
    """

    def __init__(self, factory, name, finalizer=None):
        """
        Create a session object
        """
        super().__init__(factory, name, finalizer)
        self.request_q = janus.Queue()
        self.result_q = janus.Queue()
        self.thread = None
        self.event_loop_task = None

    async def connect(self, **kwargs):
        """
        Initiate a connection
        """
        self.thread = threading.Thread(name=self.__class__, daemon=True,
                                       target=self._thread_main,
                                       kwargs=kwargs)
        self.thread.start()
        self.event_loop_task = events.create_checked_task(self._event_loop())

    async def disconnect(self):
        """
        Request a disconnection
        """
        await self._set_state(ConState.DISCONNECTING)
        await self._request("disconnect", override=True)

        # Give the thread a few seconds, then try to mop things up
        await asyncio.sleep(5)
        if self.state != ConState.DISCONNECTED:
            log.info('Cancelling {} task for slow disconnection'.format(self.name))
            state = ConState.FAILURE_WHILE_DISCONNECTING
            state.failure_reason = 'Disconnect timeout'
            await self._set_state(state)
        self.event_loop_task.cancel()
        self.event_loop_task = None

    async def _event_loop(self):
        # Permanently wait for any responses to requests that are provided
        # by the worker thread, and dispatch the responses back appropriately
        while True:
            try:
                result, fut = await self.result_q.async_q.get()
                if isinstance(result, ConState):
                    await self._set_state(result)
                elif isinstance(result, Exception):
                    fut.set_exception(result)
                else:
                    fut.set_result(result)
            except Exception as e:
                log.error('Exception during connection/threaded/_event_loop: {}'
                          ' (see debug.log for details)'.format(e))
                log.debug('_event_loop exception details', exc_info=True, stack_info=True)

    async def _request(self, action, override, *args):
        """
        Queue a request to the worker thread
        """
        # The override flag is intended for two purposes:
        #
        # - forcing a manual disconnect from any state
        # - allowing a finalizer to do operations on a newly minted connection
        #   before regular clients can do so
        #
        # However, if the flag is set, we just go and try it anyway. So if
        # it's set for a purpose that might cause an exception, you're going
        # to get an exception.
        if self.state != ConState.CONNECTED and not override:
            raise ConnectionError('Connection {} in state {} so cannot {}({})'.format(
                                self.name, self.state.name, action, args))
        fut = asyncio.Future()
        await self.request_q.async_q.put((action, args, fut))
        await asyncio.wait([fut], return_when=asyncio.ALL_COMPLETED)
        return fut.result()

    def _update_state(self, state, failure_reason=None):
        """
        Push a connection state update to the main thread
        """
        if failure_reason is not None:
            state.failure_reason = failure_reason
        self.result_q.sync_q.put((state, None))

    def _thread_main(self, **kwargs):
        """
        Kick off a session in the new thread
        """
        self.terminate = False
        initiate_connect = True
        self._update_state(ConState.CONNECTING)

        # Event loop for this session
        while not self.terminate:
            # Connect if required
            if initiate_connect:
                try:
                    self._handle_connect(**kwargs)
                except Exception as reason:
                    err = str(reason)
                    log.error('Exception in _handle_connect: ' + err + ' (see debug.log for details)')
                    log.debug('Exception details', exc_info=True, stack_info=True)
                    m = re.search('<.*>: *(.+)', err)
                    failure_reason = m.group(1) if m else err
                    self._update_state(ConState.FAILED_TO_CONNECT,
                                       failure_reason=failure_reason)
                initiate_connect = False

            # Block for a request
            action, args, fut = self.request_q.sync_q.get()
            if not (action == 'recv' and len(args) == 1 and args[0] == 0):
                log.debug('{} worker thread req: {}{}'.format(self.name, action, args))

            # Do the request
            handler = getattr(self, '_handle_' + action)
            try:
                result = handler(*args)
            except Exception as e:
                # Err on the side of caution for customer demo purposes -
                # ditch the whole thing lazily (leaking all sorts of stuff) and
                # blindly reconnect. We could instead just return the error
                # and carry on - probably more correct once we have confidence
                # that things generally do the right thing.
                failure_reason = "Handler for {}({}) crashed: {}".format(
                                                            action, args, e)
                log.warning(failure_reason + ' (see debug.log for details)')
                log.debug('Exception details', exc_info=True, stack_info=True)
                self._update_state(ConState.RECONNECTING_AFTER_FAILURE,
                                   failure_reason=failure_reason)
                initiate_connect = True
                result = e

            # Return the result
            self.result_q.sync_q.put((result, fut))

        log.debug('{} thread main exit'.format(self.name))



# Event-handling utilities
#
# Copyright (c) 2018 Ensoft Ltd

"""Event-handling utilities."""

__all__ = (
    "create_checked_task",
)


import asyncio


def create_checked_task(coro_or_future):
    """
    Wrapper for `asyncio.ensure_future` that always propagates exceptions.
    """
    def raise_any_exc(task):
        return task.result()
    task = asyncio.ensure_future(coro_or_future)
    task.add_done_callback(raise_any_exc)
    return task




# DynamicFeature base class
#
# Copyright (c) 2018 Ensoft Ltd

from .base import Feature

class DynamicFeature(Feature):
    """
    A feature that is started only when the client requests it, and thus can
    be instantiated more than once per channel (with distinct target values)
    """
    def __init__(self, ws_handler, channel, target, original_request):
        """
        Remember our channel, target, and original request (which contains
        any other parameters specific to particular subclasses)
        """
        super().__init__(ws_handler)
        self.channel = channel
        self.target = target
        # Parent target is which target we should flock under - same as target
        # for everything other than connection groups
        self.parent_target = target
        self.original_req = original_request

    async def _notify(self, **nfn):
        """
        Insert the target and channel into any outgoing notifications
        """
        nfn['channel'] = self.channel
        nfn['target'] = self.target
        await super()._notify(**nfn)

    @classmethod
    def find(cls, name):
        """
        Look up a concrete subclass by "name" field
        """
        return _find(cls, name)

def _find(cls, name):
    """
    Recursively find the named subclass
    """
    for c in cls.__subclasses__():
        if c.name == name:
            return c
        v = _find(c, name)
        if v is not None:
            return v



# Overall handler for a single websocket
#
# Copyright (c) 2018 Ensoft Ltd

from collections import defaultdict
import asyncio, logging, re, traceback
import ujson
from websockets.exceptions import ConnectionClosed
from .connection import ConState
from .feature import *
from ._util import events

log = logging.getLogger(__name__)

# Once: normalize the Feature schemas, so that subclasses pick up the supported
# requests or notifications from their ancestor classes
Feature.normalize_schema()

# Turn multiple args into a single flat key for dict lookups
def _mktuple(*args): return '||'.join(args)

class WebsocketHandler():
    """
    Class that holds a set of Feature modules, and handles the mux/demux so
    that they all share a single websocket in a sensible fashion. One of these
    objects is instantiated for each client session.
    """
    def __init__(self, ws, feature_config):
        self.ws = ws
        self.conn_factory = None
        self.request_map_default = {}
        self.request_map_optional = {}
        self.features = []
        self.target_features = defaultdict(list)
        self.target_group = {}
        self.con_state_listeners = []

        # Start out with just the configured features
        for feature_cls in ConfiguredFeature.all():
            name = feature_cls.name
            if name in feature_config:
                log.debug('Adding configured feature ' + name)
                self.add_feature(feature_cls(self, feature_config[name]))
            else:
                log.debug('Skipping unconfigured feature ' + name)

    async def handle_incoming_requests(self):
        """
        Mini-event loop that listens for incoming requests and handles them
        """
        while True:
            got_req = False
            try:
                req = await self.ws.recv()
                got_req = True
            except (asyncio.CancelledError, ConnectionClosed):
                log.info('Websocket closed')
                for feature in self.features:
                    feature.close()
                break
            except Exception as e:
                log.error('Websocket recv exception: %s', e)
            try:
                if got_req:
                    await self._handle_req(req)
            except Exception as e:
                log.error('Exception during _handle_req: {} (see debug.log for details)'.format(e))
                log.debug('_handle_req exception details', exc_info=True, stack_info=True)
                await self.notify_error(str(e))

    async def _handle_req(self, raw_request):
        """
        Handle an incoming request by dispatching to the appropriate Feature
        """
        request = ujson.loads(raw_request)
        request['userid'] = 'default' # no auth features yet
        req_type = request['req_type']
        channel = request['channel']
        target = request.get('target', '')
        if req_type != 'ping':
            log.debug('WS RECV: {}'.format(abbreviate(request)))

        # Dispatch the request
        try:
            # First try default features - these are keyed just off req_type,
            # and are executed synchronously (since they are supposed to be
            # quick, and can include meta-operations like starting new features)
            feature = self.request_map_default[req_type]
            await feature.handle(request)
        except KeyError:
            # Fall back to optional features - these are keyed off the
            # <req_type, channel, target> triple, and are executed
            # asynchronously, to avoid head-of-line blocking in complex
            # message processing
            try:
                key = _mktuple(req_type, channel, target)
                feature = self.request_map_optional[key]
                events.create_checked_task(feature.handle(request))
            except KeyError:
                log.warning('Un-handleable request {}'.format(request))
                log.debug('key = {}, request_map_optional = {}'.format(
                            key, self.request_map_optional))
                await self.notify_error("Don't know how to handle request {}".
                                        format(request))
                return

    async def notify(self, **nfn):
        """
        Send a specific outbound notification
        """
        if nfn['nfn_type'] != 'pong':
            log.debug('WS SEND: {}'.format(abbreviate(nfn)))
        json = ujson.dumps(nfn)
        await self.ws.send(json)

    async def notify_error(self, error, **nfn):
        """
        Send an outbound error, caught by the frontend's app top level
        """
        await self.notify(channel='error', nfn_type='error', value=error, **nfn)

    def add_feature(self, feature, channel=None, target=None):
        """
        Add a feature instance
        """
        self.features.append(feature)
        if isinstance(feature, ConfiguredFeature):
            # Configured feature: just add the handled request_types to the
            # default request map
            for req_type in feature.requests.keys():
                self.request_map_default[req_type] = feature
        else:
            # Dynamic feature: add the <request_type, channel, target> triple
            # to the optional request map
            assert isinstance(feature, DynamicFeature)
            for req_type in feature.requests.keys():
                key = _mktuple(req_type, channel, target)
                self.request_map_optional[key] = feature

            if isinstance(feature, TargetGroupFeature):
            # Remember target groups
                self.target_group[feature.target] = feature

            # All target features need their own collection
            if isinstance(feature, TargetFeature):
                parent_target = feature.parent_target
                if parent_target is not None:
                    self.target_features[parent_target].append(feature)
                    if parent_target in self.target_group:
                        self.target_group[parent_target].add_feature(feature)

    async def remove_feature(self, feature, channel, target):
        """
        Remove a dynamic feature instance
        """
        assert isinstance(feature, DynamicFeature)

        # Remove from the request map
        for req_type in feature.requests.keys():
            key = _mktuple(req_type, channel, target)
            del self.request_map_optional[key]

        if isinstance(feature, TargetGroupFeature):
            # Forget from set of target groups
            del self.target_group[feature.target]

            # The dying feature's children will just forget about it. Harsh.
            for child in feature.children:
                child.parent_target_group = None

        # Same for the dying feature's parents. Even harsher.
        if isinstance(feature, TargetFeature):
            parent_target_group = feature.parent_target_group
            if parent_target_group is not None:
                # First consider the dying feature to be disconnected, so the
                # parent feature recomputes its overall state
                saved_state = feature.state     # bit of a hack
                feature.state = ConState.DISCONNECTED
                await parent_target_group.state_listener(feature)
                feature.state = saved_state      # at least hack over now

                # Then actually make your parents forget you, Hermione
                parent_target_group.remove_feature(feature)
            if feature.parent_target is not None:
                self.target_features[feature.parent_target].remove(feature)


    def get_features_for_target(self, target):
        """
        Return all target features for a given target
        """
        return self.target_features[target]


MAX_LENGTH = 200
def abbreviate(msg):
    """
    Helper function to make logging safer and saner
    """
    val = msg.copy()
    if isinstance(val, dict):
        for k, v in val.items():
            if k == 'secret' and isinstance(v, str):
                val[k] = '*' * len(v)
            elif isinstance(v, str) and len(v) > MAX_LENGTH:
                val[k] = v[0:MAX_LENGTH] + "..."
            elif isinstance(v, dict):
                val[k] = abbreviate(v)
    return val



# Persistence feature
#
# Copyright (c) 2018 Ensoft Ltd

import ujson
from collections import defaultdict
from .cfg_base import ConfiguredFeature

# Remember which active websockets have requested data for a given channel, so
# that any changes can get published to everyone.
# userid name -> channel name -> set of interested PersistFeature instances
listeners = defaultdict(lambda: defaultdict(set))

class PersistFeature(ConfiguredFeature):
    """
    Feature that saves and retrieves arbitrary data chunks for the frontend app.
    Intended for small bits of data (eg preferences) - load/save operations
    are not expected to be efficient. This assumes it is called only from
    one thread (the main event loop thread). Each channel (for each userid)
    can save or load one JSONable value.
    """
    #
    # Schema
    #
    name = 'persist'

    requests = {'persist_save_async': ['userid', 'channel', 'data'],
                'persist_save_sync':  ['userid', 'channel', 'data'],
                'persist_load':       ['userid', 'channel', 'default']}

    config = {'filename': 'persist.json'}

    # Unsubscribe ourselves from everything on close
    def close(self):
        for userid in listeners.values():
            for channels in userid.values():
                channels.remove(self)

    #
    # Implementation
    #
    async def do_persist_save_async(self, userid, channel, data):
        """
        Save a table, overwriting it if already present
        """
        db = self._load_db()
        if userid not in db:
            db[userid] = {channel: data}
        else:
            db[userid][channel] = data
        self._save_db(db)

        # Notify any other peer connections that care about this
        for obj in listeners[userid][channel]:
            if obj != self:
                await obj._notify(nfn_type='persist_load',
                                   channel=channel, data=data)

    async def do_persist_save_sync(self, userid, channel, data):
        """
        Same as do_persist_save, with a synchronous reply.
        """
        try:
            await self.do_persist_save_async(userid, channel, data)
            return self._rpc_success('')
        except Exception as e:
            return self._rpc_failure(e)

    async def do_persist_load(self, userid, channel, default):
        """
        Load a table. If not present then return the specified default
        """
        listeners[userid][channel].add(self) # subscribe
        db = self._load_db()
        try:
            data = db[userid][channel]
        except KeyError:
            data = default
        return self._result('persist_load', data=data)

    def _load_db(self):
        """
        Load the entire database
        """
        try:
            # tsk - synchronous file I/O. ho hum.
            with open(self.config['filename']) as f:
                return ujson.loads(f.read())
        except FileNotFoundError:
            return {}
        except ValueError as e:
            raise Exception("{} is invalid json: {}".format(
                self.config['filename'], e))

    def _save_db(self, db):
        """
        Save the entire database
        """
        # tsk - synchronous file I/O again. la di da.
        with open(self.config['filename'], 'w') as f:
            f.write(ujson.dumps(db, indent=4))



# Target group feature
#
# Copyright (c) 2018 Ensoft Ltd

import asyncio, time

from .._util import events
from .tgt_base import TargetFeature

class TargetGroupFeature(TargetFeature):
    """
    Feature that manages a set (or subtree) of target features
    """
    #
    # Schema
    #
    name = 'target_group'

    #
    # Implementation
    #
    def __init__(self, ws_handler, channel, target, original_request):
        """
        Initialize ourselves for the specified target
        """
        super().__init__(ws_handler, channel, target, original_request)
        self.parent_target = original_request.get('parent_target', None)

        # If there are target features for our member target that have
        # already been started, associate them with us now. Any ones created
        # subsequently will automatically invoke our add_feature() method.
        for feature in ws_handler.get_features_for_target(target):
            assert feature != self
            self.add_feature(feature)

    async def connect(self, conn_factory):
        """
        Initiate all the connections to this target
        """
        for child in self.children:
            events.create_checked_task(child.connect(conn_factory))

    def add_feature(self, feature):
        """
        Add a target feature for our target
        """
        self.children.add(feature)
        feature.parent_target_group = self
        if self.connect_requested:
            # A connect request has already been made. So the late-arrival
            # feature should try to connect now too.
            events.create_checked_task(feature.connect(self.conn_factory))

    def remove_feature(self, feature):
        """
        Remove a target feature from our target
        """
        self.children.remove(feature)



# Syslog monitoring feature
#
# Copyright (c) 2018 Ensoft Ltd

import asyncio, re, time

from .._util import events
from .tgt_base import TargetFeature

class SyslogFeature(TargetFeature):
    """
    Feature that monitors for syslogs, with optional debugs and filtering.
    """
    #
    # Schema
    #
    name = 'syslog'
    notifications = ['syslog']

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Set up our connection to the router
        """
        self.connection = await conn_factory.get_cli_connection(
                                'syslog', finalizer=self.finalizer)
        self.add_connection(self.connection, from_scratch=True)

    async def finalizer(self):
        """
        Finalize a new connection
        """
        # First handle any filters/debug settings from the start_feature req
        filters = self.original_req.get('filters', [])
        if len(filters) == 0:
            # just drop empty lines
            filters = ['\S']
        regexp = re.compile('|'.join(['({})'.format(f) for f in filters]))
        nfn = {'nfn_type': 'syslog',
               'channel': self.original_req['channel']}
        if 'id' in self.original_req:
            nfn['id'] = self.original_req['id']

        # Then do the expect stuff to get us ready to go
        await self.connection.send('undebug all all-tty\n', override=True)
        await self.connection.expect_prompt(override=True)
        await self.connection.send('terminal monitor\n', override=True)
        await self.connection.expect_prompt(override=True)
        for debug in self.original_req.get('debugs', []):
            await self.connection.send('{}\n'.format(debug), override=True)
            await self.connection.expect_prompt(override=True)

        # We need to return at this point, so the connection transitions
        # from FINALIZING to CONNECTED, so tee up our mini event loop for later
        events.create_checked_task(self._event_loop(regexp, nfn))

    async def _event_loop(self, regexp, nfn):
        """
        Sit and wait for syslogs/debugs to come in
        """
        # poll event loop for disconnect requests occasionally
        await self.connection.settimeout(1)

        # enter event loop
        while not self.connection.terminate:
            data = (await self.connection.recv()).decode()
            for syslog in data.split('\n'):
                if regexp.search(syslog):
                    nfn['result'] = syslog
                    nfn['time'] = time.time()
                    await self._notify(**nfn)



# EnTrance default application start
#
# Copyright (c) 2018 Ensoft Ltd

import argparse, logging, logging.config, sys
import sanic, sanic.response, yaml

from . import WebsocketHandler
from ._util import logger

log = logging.getLogger(__name__)
location = getattr(sys, '_MEIPASS', '.') + '/' # where pre-canned files live

def start(config, task):
    """
    Start a simple server with the specified configuration. If an app needs
    more elaborate setup, then just copy this function and modify.
    """
    start_cfg = config['start']
    log.info('Starting app with ' +
             ', '.join(['{}={}'.format(k, v) for k, v in start_cfg.items()]))
    app = sanic.Sanic(log_config=None)
    app.config.RESPONSE_TIMEOUT = 3600
    app.config.KEEP_ALIVE_TIMEOUT = 75

    # Websocket handling
    @app.websocket('/ws')
    async def handle_ws(request, ws):
        log.info('New websocket client')
        ws_handler = WebsocketHandler(ws, config['features'])
        await ws_handler.handle_incoming_requests()

    # Static file handling
    static_dir = location + start_cfg['static_dir']
    app.static('/', static_dir)
    @app.route('/')
    async def home_page(request):
        return await sanic.response.file(static_dir + '/index.html')

    # Optionally invoke a caller-specified task once the event loop is up
    if task is not None:
        app.add_task(task)

    # Enter event loop
    app.run(host=start_cfg['host'], port=start_cfg['port'])

def parse_args(args):
    """
    Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--port',
                        help='override port')
    parser.add_argument('-a', '--addr',
                        help='override bind address')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='debug to console')
    parser.add_argument('-c', '--config', default=location + 'config.yml',
                        help='config file')
    parser.add_argument('-l', '--logging',
                        help='logging definition file')
    return parser.parse_args(args)

def main(*args, task=None):
    # Load up preferences
    opts = parse_args(args)

    # Config file is always required. The logging file may fall back
    # to the default yaml file below if unspecified.
    main_config = yaml.safe_load(open(opts.config).read())
    if opts.logging is None:
        logging_yaml = logging_yaml_default
    else:
        logging_yaml = open(opts.logging).read()
    logging_config = yaml.safe_load(logging_yaml)

    # Apply any command-line overrides
    if opts.port is not None:
        main_config['start']['port'] = opts.port
    if opts.addr is not None:
        main_config['start']['host'] = opts.addr
    if opts.debug:
        logging_config['handlers']['console']['level'] = 'DEBUG'

    # Go
    logging.config.dictConfig(logging_config)
    logging.setLogRecordFactory(logger.FormattedLogRecord)
    start(main_config, task)
    log.info('Closing down gracefully')

# Default logging.yml file
logging_yaml_default="""
# Separate configuration file for logging, since most end-users care less
# about the details of this
version: 1
disable_existing_loggers: false
formatters:
    brief:
        format: "%(levelname)s: [%(name)s] %(message)s"
    normal:
        format: "%(asctime)s %(levelname)s: [%(name)s] %(message)s"
handlers:
    console:
        class: logging.StreamHandler
        level: INFO
        formatter: brief
        stream: ext://sys.stdout
    debug:
        class: logging.handlers.RotatingFileHandler
        level: DEBUG
        formatter: normal
        filename: debug.log
        maxBytes: 1000000
        backupCount: 3
        encoding: utf8
loggers:
    "":
        level: DEBUG
        handlers: [console, debug]
    asyncio:
        level: DEBUG
        handlers: [debug]
    paramiko:
        level: WARNING
        handlers: [console, debug]
    ncclient:
        level: WARNING
        handlers: [console, debug]
    network:
        level: DEBUG
        handlers: [debug]
    sanic:
        level: WARNING
        handlers: [debug]
    websockets:
        level: WARNING
        handlers: [console, debug]
"""

if __name__ == '__main__':
    main(*sys.argv[1:])



# Maintain a persistent ssh connection to an individual router
#
# Copyright (c) 2018 Ensoft Ltd

import os, re, socket, sys, time
try:
    import paramiko
    import ncclient.manager as nc_mgr
    from ncclient.operations import RaiseMode
    from entrance.connection.base import ConnectionFactory, ConState
    from entrance.connection.cli import ThreadedCLIConnection
    from entrance.connection.netconf import ThreadedNCConnection
except ImportError:
    # Very cheesy way to permit the majority case (router interaction
    # features not required) to install the entrance package without
    # incurring all the complex dependencies from the router features.
    # Better solutions welcomed!
    class Fail():
        def __init__(self, *args, **kwargs):
            print('\n\n\n\n* To use router features, re-install depending',
                  'on the package\n* name "entrance[with-router-features]",',
                  'not simply "entrance".\n* This installation does not',
                  'have the required dependencies.\n\n\n', file=sys.stderr)
            os.abort()
    ThreadedCLIConnection = Fail
    ThreadedNCConnection = Fail
    ConnectionFactory = Fail

__all__ = ['SSHConnectionFactory']

BUF_SIZE = 10000 # ssh max buffer size

class SSHCLIConnection(ThreadedCLIConnection):
    """
    SSH CLI Connection
    """
    def _handle_connect(self, **creds):
        """
        Initiate a persistent ssh connection, in the paramiko thread.
        """
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        kwargs = {'username': creds['username'],
                  'port': int(creds.get('ssh_port', '22')),
                  'allow_agent': False, 'look_for_keys': False}
        if 'ssh_key' in creds:
            kwargs['key_filename'] = creds['ssh_key']
        else:
            kwargs['password'] = creds.get('password', '')
        self.ssh.connect(creds['host'], **kwargs)
        self.channel = self.ssh.invoke_shell(width=0, height=0)

        # Swallow any initial stuff
        self.channel.send('\n')
        while True:
            z = self.channel.recv(BUF_SIZE).decode()
            if '#' in z:
                break
            time.sleep(1)

        # Should be sane now!
        self._update_state(ConState.FINALIZING)

    def _handle_settimeout(self, timeout):
        """
        Set a timeout on send/recv requests
        """
        self.channel.settimeout(timeout)

    def _handle_disconnect(self):
        """
        Kill the connection
        """
        try:
            self.ssh.close()
            self._update_state(ConState.DISCONNECTED)
        except Exception as e:
            self._update_state(ConState.FAILURE_WHILE_DISCONNECTING, str(e))
        self.terminate = True

    def _handle_send(self, data):
        """
        Send some data
        """
        self.channel.send(data)

    def _handle_recv(self, nbytes):
        """
        Wait for some data
        """
        try:
            if nbytes > 0:
                buf = self.channel.recv(nbytes)
            else:
                buf = self.channel.recv(BUF_SIZE)
                while self.channel.recv_ready():
                    buf += self.channel.recv(BUF_SIZE)
            return buf
        except socket.timeout:
            return bytes()

class SSHNCConnection(ThreadedNCConnection):
    """
    SSH Netconf Connection
    """
    def _handle_connect(self, **creds):
        """
        Initiate a persistent ssh connection, in the worker thread
        """
        kwargs = {'username': creds['username'],
                  'port': creds.get('netconf_port', 830),
                  'device_params': {'name':'iosxr'}}
        if 'ssh_key' in creds:
            kwargs['key_filename'] = creds['ssh_key']
        else:
            kwargs['password'] = creds.get('password', '')
        try:
            self.mgr = nc_mgr.connect_ssh(creds['host'], **kwargs)
            self.mgr.raise_mode = RaiseMode.NONE
            self._update_state(ConState.FINALIZING)
        except Exception as e:
            self.mgr = None
            self._update_state(ConState.FAILED_TO_CONNECT, str(e))

    def _handle_disconnect(self):
        """
        Kill the connection
        """
        try:
            if self.mgr is not None:
                self.mgr.close_session()
            self._update_state(ConState.DISCONNECTED)
        except Exception as e:
            self._update_state(ConState.FAILURE_WHILE_DISCONNECTING, str(e))
        self.terminate = True

class SSHConnectionFactory(ConnectionFactory):
    """
    ConnectionFactory for a regular router
    """
    cli_connection_cls = SSHCLIConnection
    netconf_connection_cls = SSHNCConnection



# Feature base class
from .base import Feature

# Configured features
from .cfg_base import ConfiguredFeature
from .cfg_core import CoreFeature
from .cfg_persist import PersistFeature

# Dynamic base class
from .dyn_base import DynamicFeature

# Target features
from .tgt_base import TargetFeature
from .tgt_group import TargetGroupFeature

from .tgt_cli_config import CLIConfigFeature
from .tgt_cli_exec import CLIExecFeature
from .tgt_netconf import NetconfFeature
from .tgt_syslog import SyslogFeature



# ConnectionFeature base class
#
# Copyright (c) 2018 Ensoft Ltd

import asyncio, logging, time

from ..connection import connection_factory_by_name, ConState, Connection
from .._util import events
from .dyn_base import DynamicFeature

log = logging.getLogger(__name__)

class TargetFeature(DynamicFeature):
    """
    A feature that has one or more connections to a target (eg a router). Always
    optional, because we don't want connection initiation unless needed by the
    application.
    """
    #
    # Schema
    #
    requests = {'connect': ['connection_type', 'params'],
                'disconnect': []}

    notifications = ['connection_state']

    #
    # Implementation
    #
    def __init__(self, ws_handler, channel, target, original_request):
        """
        Initialize the connections we actively own.
        """
        super().__init__(ws_handler, channel, target, original_request)
        self.children = set()
        self.state = ConState.DISCONNECTED
        self.parent_target_group = None
        self.connect_requested = False

        # If the client has requested connection state updates, then set up
        # the common part of each notification message now
        if original_request.get('con_state_subscribe', False):
            self.state_subscription_nfn = {
                'nfn_type': 'connection_state',
                'feature': self.name
            }
            if 'id' in original_request:
                self.state_subscription_nfn['id'] = original_request['id']
        else:
            self.state_subscription_nfn = None

    async def do_connect(self, connection_type, params):
        """
        Create a connection factory and start connecting
        """
        self.connect_requested = True
        if params.get('auth_is_password', True):
            params['password'] = params['secret']
        else:
            params['ssh_key'] = params['secret']
        conn_factory_cls = connection_factory_by_name[connection_type]
        self.conn_factory = conn_factory_cls(**params)
        # Actually do it
        await self.connect(self.conn_factory)

    async def do_disconnect(self):
        """
        Disconnect from remote protocol peers
        """
        # We shouldn't be asked to disconnect if it wasn't us who connected
        # originally
        if not self.connect_requested:
            log.info("{} disconnecting, although we didn't connect originally"
                     .format(self.name))
        self.connect_requested = False
        # Actually do it
        await self.disconnect()

    async def connect(self, conn_factory):
        """
        Actually kick off the outgoing connections - implemented by subclasses
        """
        assert False

    async def disconnect(self):
        """
        Disconect
        """
        # Connections can be removed mid-iteration if they disconnect promptly
        safe_iter = list(self.children)
        for child in safe_iter:
            events.create_checked_task(child.disconnect())

    def add_connection(self, connection, from_scratch=False):
        """
        Associate a Connection object with this Feature object, for the purposes
        of global connection state monitoring
        """
        if from_scratch:
            self.children = set()
        self.children.add(connection)
        connection.add_state_listener(self.state_listener)

    async def state_listener(self, child):
        """
        Callback when one of our feature's connections changes state
        """
        if child not in self.children:
            # Possible race condition - a dying connection is still talking
            # to us, but we don't care any more
            return

        # Calculate aggregate state across all child connections/features
        self.state = child.state
        for c in self.children:
            if c.state > self.state:
                self.state = c.state

        # Send the notification if we were asked to
        if self.state_subscription_nfn is not None:
            def encode_state(state):
                return {
                    'state': state.name,
                    'error': state.failure_reason if state.is_failure() else ''
                }
            nfn = self.state_subscription_nfn.copy()
            nfn['child'] = child.name
            nfn['child_state'] = encode_state(child.state)
            nfn['feature'] = self.name
            nfn['state'] = encode_state(self.state)
            nfn['state_is_up'] = self.state == ConState.CONNECTED
            nfn['timestamp'] = time.strftime('%H:%M:%S')
            await self._notify(**nfn)

        # Independently notify our target manager, if there is one
        if self.parent_target_group is not None:
            await self.parent_target_group.state_listener(self)

        # If a connection object has disconnected, then throw it away - we'll
        # create a new one if there's another connection request later
        if child.state == ConState.DISCONNECTED and isinstance(child, Connection):
            self.children.remove(child)



# Netconf feature
#
# Copyright (c) 2018 Ensoft Ltd

from .tgt_base import TargetFeature

class NetconfFeature(TargetFeature):
    """
    Feature that exposes Netconf operations
    """
    #
    # Schema
    #
    name = 'netconf'
    requests = {'netconf': ['op', '__req__']}

    # We decode out the actual netconf op as follows
    op_requests = {'get': ['value'],
                   'get_config': ['value'],
                   'edit_config': ['value'],
                   'commit': [],
                   'validate': [],
                   'discard_changes': []}

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Connect and get ready for the next commit request
        """
        self.connection = await conn_factory.get_netconf_connection('netconf')
        self.add_connection(self.connection, from_scratch=True)

    async def do_netconf(self, op, req):
        """
        Do one of the netconf operations specified in op_requests
        """
        try:
            fn = getattr(self.connection, op)
            args = [req[arg] for arg in self.op_requests[op]]
            rpc_reply = await fn(*args)
            result = str(rpc_reply)

            # Semi-hack for commit check:
            #
            # If a validate operation returns errors, but they all have
            # severity 'warning' not 'error', then consider the operation
            # a success
            if op == 'validate' and not rpc_reply.ok and \
               not '<error-severity>error</error-severity>' in result:
               return self._rpc_success(result)

            # Usual processing
            elif rpc_reply.ok:
                return self._rpc_success(result)
            else:
                return self._rpc_failure(result)

        except Exception as e:
            return self._rpc_failure(str(e))



# EnTrance logging utils
#
# Copyright (c) 2018 Ensoft Ltd

"""Logging helpers and utils."""

__all__ = (
    "FormattedLogRecord"
)


import logging


class FormattedLogRecord(logging.LogRecord):
    """
    Log record that supports "{"-style formatting.

    Can be passed to `logging.setLogRecordFactory` to install it as the
    default log record type in the `logging` package.

    """
    def getMessage(self):
        """Return the message for this record."""
        msg = str(self.msg)
        # Short-circuit for old-style formatting.
        if "%" in msg:
            return super().getMessage()
        if self.args:
            # Three possibilities (which match the three
            #  - 'args' is a simple dictionary (on its own) that is to be
            #    output in full in a single "{}". If "format(*self.args)" is
            #    used then only the first item in the dictionary is output.
            #  - The format string doesn't have naming parameters.
            #  - The format string does have naming parameters and so
            #    format_map() must be used.
            try:
                if isinstance(self.args, dict):
                    msg = msg.format(self.args)
                else:
                    msg = msg.format(*self.args)
            except KeyError:
                msg = msg.format_map(self.args)
        # Now truncate the message, if necessary
        truncation = getattr(self, "truncation", None)
        if truncation is not None and len(msg) > truncation:
            msg = "".join((msg[:truncation - 3], "..."))
        return msg




# Base class for Connection and ConnectionFactory
#
# Copyright (c) 2018 Ensoft Ltd

import asyncio
import logging
from enum import IntEnum, unique

from .._util import events


@unique
class ConState(IntEnum):
    # Order is important: overall connection state is the max of the
    # states of the individual connections
    DISCONNECTED = 0
    CONNECTED = 1
    FAILURE_WHILE_DISCONNECTING = 2
    FINALIZING = 3
    CONNECTING = 4
    DISCONNECTING = 5
    RECONNECTING_AFTER_FAILURE = 6
    FAILED_TO_CONNECT = 7

    def is_failure(self):
        return self in (self.FAILURE_WHILE_DISCONNECTING,
                        self.RECONNECTING_AFTER_FAILURE,
                        self.FAILED_TO_CONNECT)

class ConnectionFactory():
    """
    Factory that knows about a single router, and can produce Connection
    objects on demand. Each Connection object represents a connection to that
    same router, that can be used for different purposes concurrently.
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    async def get_cli_connection(self, connection_name, finalizer=None):
        """
        Create a new CLI exec connection, and return a corresponding
        CLIConnection object
        """
        con = self.cli_connection_cls(self, connection_name, finalizer)
        await con.connect(**self.kwargs)
        return con

    async def get_netconf_connection(self, connection_name, finalizer=None):
        """
        Create a new Netconf connection, and return a corresponding
        NetconfConnection object
        """
        con = self.netconf_connection_cls(self, connection_name, finalizer)
        await con.connect(**self.kwargs)
        return con

class Connection():
    """
    Base class representing a single router connection
    """

    # Internal attributes:
    #
    # _log
    #   Logger object
    #
    _log = None

    def __init__(self, factory, name, finalizer, **kwargs):
        self.factory = factory
        self.name = name
        self.finalizer = finalizer
        self.state = ConState.DISCONNECTED
        self.kwargs = kwargs
        self.state_listeners = []
        self._log = logging.getLogger(
            "{}.{}-{}".format(__name__, type(self).__name__, name))

    def add_state_listener(self, listener):
        """
        Add a callback to be notified when connection changes state
        """
        self.state_listeners.append(listener)

    async def connect(self):
        # documentation only, implemented by subclasses
        pass

    async def disconnect(self):
        # documentation only, implemented by subclasses
        pass

    async def _set_state(self, state):
        """
        Helper method to set the connection state.
        SUBCLASSES MUST ALWAYS USE THIS, DON'T JUST SET THE STATE DIRECTLY
        """
        # Actually change state and then notify listeners.
        old_state = self.state
        self.state = state
        for listener in self.state_listeners:
            await listener(self)
        self._log.debug("changed state from {} to {}; notified {} listeners",
                        old_state.name, state.name, len(self.state_listeners))

        # Call the finalizer if required
        if state == ConState.FINALIZING:
            async def finalize():
                self._log.debug("started finalizing")
                if self.finalizer is not None:
                    await self.finalizer()
                await self._set_state(ConState.CONNECTED)
                self._log.debug("finished finalizing")
            events.create_checked_task(finalize())



# CLI Exec feature
#
# Copyright (c) 2018 Ensoft Ltd

from .tgt_base import TargetFeature

class CLIExecFeature(TargetFeature):
    """
    Feature that submits CLI exec requests directly
    """
    #
    # Schema
    #
    name = 'cli_exec'
    requests = {'cli_exec': ['command']}

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Connect and get ready for future cli_exec requests
        """
        self.connection = await conn_factory.get_cli_connection(
                                                    'cli_exec', self.finalizer)
        self.add_connection(self.connection, from_scratch=True)

    async def finalizer(self):
        """
        Finalize a new connection
        """
        await self.connection.send('run stty rows 0\n', override=True)
        await self.connection.expect_prompt(override=True)

    async def do_cli_exec(self, command):
        """
        Do a single CLI exec command
        """
        await self.connection.send(command + "\n")
        output = await self.connection.expect_prompt()
        m = self.connection._interesting.search(output)
        if m:
            return self._rpc_success(m.group(1))
        else:
            return self._rpc_failure(output)



# CLI config feature
#
# Copyright (c) 2018 Ensoft Ltd

import re
from .tgt_base import TargetFeature

class CLIConfigFeature(TargetFeature):
    """
    Feature that enters configuration via the CLI
    """
    #
    # Schema
    #
    name = 'cli_config'
    requests = {'cli_config_load': ['config'],
                'cli_config_commit': ['check_only'],
                'cli_config_get_failures': [],
                'cli_config_get_unsupported': []}

    #
    # Implementation
    #
    async def connect(self, conn_factory):
        """
        Connect and get ready for the next commit request
        """
        self.connection = await conn_factory.get_cli_connection('config', finalizer=self.finalizer)
        self.add_connection(self.connection, from_scratch=True)

    async def finalizer(self):
        """
        Finalize a new connection
        """
        await self.connection.send('configure\n', override=True)
        await self.connection.expect_prompt(override=True)

    async def do_cli_config_load(self, config):
        """
        Load up a config buffer with some CLI
        """
        # Clear any previous confiuration in the session
        await self.connection.send('clear\n')
        await self.connection.expect_prompt()

        # Enter the new configuration
        errors = []
        for line in config.split('\n'):
            await self.connection.send(line + '\n')
            result = await self.connection.expect_prompt()
            if len(result) > len(line) + 4:
                errors.append(result)

        # Check for parser-rejected syntax errors
        if len(errors) > 0:
            return self._rpc_failure('\n'.join(errors))
        else:
            return self._rpc_success()

    async def do_cli_config_commit(self, check_only=False):
        """
        Commit a config buffer populated with cli_config_load
        """
        command = 'validate commit' if check_only else 'commit'
        await self.connection.send(command + '\n')
        result = await self.connection.expect_prompt()
        if len(result) > 200:
            return self._rpc_failure(result)
        else:
            return self._rpc_success()

    async def do_cli_config_get_failures(self):
        """
        Get config errors
        """
        await self.connection.send('show configuration failed\n')
        result = await self.connection.expect_prompt(strip_top=True)
        result = result.strip()
        if len(result):
            return self._rpc_failure(result)
        else:
            return self._rpc_success()

    async def do_cli_config_get_unsupported(self):
        """
        Get config items that are unsupported by validation
        """
        await self.connection.send('show configuration validation unsupported\n')
        result = await self.connection.expect_prompt(strip_top=True)
        result = result.strip()
        if result == '% No such configuration item(s)':
            return self._rpc_success()
        else:
            return self._rpc_failure(result)



# ConfiguredFeature base class
#
# Copyright (c) 2018 Ensoft Ltd

import logging, sys
from .base import Feature

log = logging.getLogger(__name__)

class ConfiguredFeature(Feature):
    """
    A feature that is started by configuration, once per websocket
    """
    # Dictionary of configuration options and default values
    config = {}

    def __init__(self, ws_handler, config):
        """
        Merge in any specifed configuration with our defaults
        """
        super().__init__(ws_handler)
        for key, val in config.items():
            if key in self.config:
                self.config[key] = val
            else:
                log.critical('\n!!\n!!\n!! Invalid config item "{}" for feature {}\n'
                          '!! Possible items are: {}\n!!\n!!\n!!'.format(
                              key, self.name, sorted(self.config.keys())))
                sys.exit(1)

    @classmethod
    def all(cls):
        """
        List all subclasses
        """
        # for now assume a flat subclass hierarchy
        return cls.__subclasses__()



# Setup definition for EnTrance package
#
# To upload a new version to PyPi using nix:
#
# - `nix-shell -p 'python3.withPackages(p:[p.pip p.setuptools p.twine])'`
# - `python3 setup.py sdist`
# - If that all looks ok, `twine upload dist/*`
#
# I suspect there's something else that should be happening with wheel, but
# this seems to work.

import os, sys
from setuptools import find_packages, setup

# Router features require heavy dependencies, so include only when actually required
# via depending on 'entrance[with-router-features' rather than just 'entrance'.
router_feature_deps = ['janus', 'ncclient', 'paramiko']

# Acutally, an icky second way of including the optional dependencies would just be
# to rewrite this from '[]' to 'router_feature_deps'. I'm looking at you, nix...
extra_deps = []

# Sanic currently requires websockets 7.x
websockets_fix = 'websockets>=7.0,<8.0'

# Sanic used to support Python 3.5+, but has recently moved on to 3.6+, with
# a long-term support release of 18.12 for 3.5
v = sys.version_info
assert v.major == 3
assert v.minor >= 5
sanic_fix = 'sanic==18.12.0' if v.minor == 5 else 'sanic'

with open('README.md', 'r') as f:
    long_description = f.read()

setup(name='entrance',
      version='1.1.6',
      author='Ensoft Ltd',
      description='Server framework for web apps',
      url='https://github.com/ensoft/entrance',

      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=find_packages(),

      classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
      ],

      install_requires=['pyyaml', sanic_fix, websockets_fix] + extra_deps,

      extras_require={
        'with-router-features': router_feature_deps
      }
)



# Base class for a CLI ThreadedConnection
#
# Copyright (c) 2018 Ensoft Ltd

import re
from entrance.connection.threaded import ThreadedConnection

class ThreadedCLIConnection(ThreadedConnection):
    """
    Base class for a ThreadedConnection whose worker thread
    maintains a CLI session
    """
    async def send(self, data, override=False):
        """
        Send some data into the connection
        """
        return await self._request('send', override, data)

    async def recv(self, nbytes=0, override=False):
        """
        Wait for some data from the connection. Note that this will cause the
        worker thread to block until some data is available. If nbytes == 0 then
        get all the data available at first shot.
        """
        return await self._request('recv', override, nbytes)

    async def settimeout(self, timeout, override=False):
        """
        Set a timeout on send/recv operations. If hit, recv will just return a
        shorter or empty response. Sends will silently drop.
        """
        return await self._request('settimeout', override, timeout)

    # Regexps for _expect_prompt below
    _prompt = re.compile(r'(.*)RP/0/(RP)?0/CPU0:[^\r\n]*?#', re.DOTALL)
    _interesting = re.compile(r'[^\n]*\n[^\n]* UTC\r\n(.*)', re.DOTALL)

    async def expect_prompt(self, strip_top=False, override=False):
        """
        Waits for a prompt, and returns all the characters up to that point
        (optionally also stripping off an initial line and timestamp)
        """
        buf = bytes()
        while True:
            buf += await self.recv(override=override)
            m = self._prompt.match(buf.decode())
            if m:
                result = m.group(1)
                if strip_top:
                    m = self._interesting.match(result)
                    if m:
                        result = m.group(1)
                return result
